#####
# HDL Generation routines
#####

# list formats used([]'s are optional):
#	library = <literal string> 
#	generic = { <name> <width> <logic_type> [<default value>] }
#		<logic_type> = 'l' = std_logic, 'u' = std_ulogic
#	port = { <name> <direction> <logic_type> <width> }
#		<direction> = 'i' = in, 'o' = out, 'io' = inout
#	component = { <name> {<generics>} {<ports>} }
#	signal = { <name> <logic_type> <width> }
#	assignment = { <sig name> < expression> }  (currently only simple a = b are implemented}
#	instances = { <name> <component_name> <generic_parameters> <port_maps> }
#		<generic_parameters> = { <generic_name> <generic_value> ... } 
#		<port_maps> = { <port_name> <signal_name> }
namespace eval ::hdl_gen {
	namespace eval fileOutput {
		namespace export *
		proc writeLine {fd tabLvl str} {
			puts $fd "[string repeat {   } $tabLvl]$str"
		}
	}

	namespace eval vhdl {
		namespace export * 
		namespace import [namespace parent]::fileOutput::*
		
		proc genPort {port} {
			set name [lindex $port 0]
			set dir [lindex $port 1]
			set logicType [lindex $port 2]
			set width [lindex $port 3]
			
			switch $dir {
				{i} {set dirStr "in"} 
				{o} {set dirStr "out"}
				{io} {set dirStr "inout"}
				default {set dirStr "Unknown:$dir"} 
			}
			switch $logicType {
				{s} {set logicTypeStr "std_logic" }
				{u} {set logicTypeStr "std_ulogic"}
				default {set logicTypeStr "Unknown:$logicType"} 
			}
			if {$width > 0} {
				set widthStr "_vector([expr $width - 1] downto 0)"
			} else {
				set widthStr ""
			}
			
			return "$name : $dirStr $logicTypeStr$widthStr"
		}
		
		
		proc writeArch {fd archVarIn} {
			upvar $archVarIn archVar
			set tabLvl 0

			#libraries
			foreach lib $archVar(libraries) {
				writeLine $fd $tabLvl "$lib"
			}
			writeLine $fd 0 ""
			# entity
			writeLine $fd $tabLvl "entity $archVar(name) is" 
			incr tabLvl
			# generics
			# ports
			writeLine $fd $tabLvl "port ("
			incr tabLvl
			set firstPorts [lrange $archVar(ports) 0 end-1 ]
			set lastPort [lindex $archVar(ports) end]
			foreach port $firstPorts {
				writeLine $fd $tabLvl "[genPort $port];"
			}
			writeLine $fd $tabLvl "[genPort $lastPort]"
			incr tabLvl -1
			writeLine $fd $tabLvl ");"
			incr tabLvl -1
			writeLine $fd $tabLvl "end;"
			
			# architecture
			writeLine $fd $tabLvl "architecture $archVar(model) of $archVar(name) is" 
			incr tabLvl
			# components
			# signals
            incr tabLvl -1
			writeLine $fd $tabLvl "begin"
			incr tabLvl
			# assignments
			# instances
			incr tabLvl -1
            writeLine $fd $tabLvl "end;"
		}
	}
	namespace eval verilog {
	}
	
	# create and return an empty architecture array (which is an array of lists of lists)
	proc createArch {archVar name model} {
		set varList "name $name 
						model $model 
						libraries {}
						generics {}
						ports {}
						components {}
						signals {}
						assignments {}"
		uplevel 1 [list array set $archVar $varList]
	}
	proc addLibrary {archVarIn library } {
		upvar $archVarIn archVar
		set archVar(libraries) [linsert $archVar(libraries) end "$library"]
	}
	proc addPort {archVarIn name direction logicType width } {
		upvar $archVarIn archVar
		set archVar(ports) [linsert $archVar(ports) end [list $name $direction $logicType $width]]
	}
	proc addSignal {archVarIn name logicType width } {
		upvar $archVarIn archVar
		set archVar(signals) [linsert $archVar(signals) end [list $name $logicType $width]]
	}

}
