//////////////////////////////////////////////////////////////////////////
// Copyright (c) 2002-2007 by Impulse Accelerated Technologies, Inc.    //
// All rights reserved.                                                 //
//                                                                      //
// This source file may be used and redistributed without charge        //
// subject to the provisions of the IMPULSE ACCELERATED TECHNOLOGIES,   //
// INC. REDISTRIBUTABLE IP LICENSE AGREEMENT, and provided that this    //
// copyright statement is not removed from the file, and that any       //
// derivative work contains this copyright notice.                      //
//////////////////////////////////////////////////////////////////////////


module fneg_ll(a, result);
	input [32-1:0] a;
	output [32-1:0] result;

	assign result[31] = a[31] ^ 1;
	assign result[30:0] = a[30:0];
endmodule

module fabs_ll(a, result);
	input [32-1:0] a;
	output [32-1:0] result;

	assign result[31] = 0;
	assign result[30:0] = a[30:0];
endmodule

module fnegd_ll(a, result);
	input [64-1:0] a;
	output [64-1:0] result;

	assign result[63] = a[63] ^ 1;
	assign result[62:0] = a[62:0];
endmodule

module fabsd_ll(a, result);
	input [64-1:0] a;
	output [64-1:0] result;

	assign result[63] = 0;
	assign result[62:0] = a[62:0];
endmodule


module ftod_ll(clk, a, result);
	input clk;
	input [32-1:0] a;
	output [64-1:0] result;

	wire [7:0] exp_s;
	wire [10:0] exp_d;
	wire zero;
  
	assign exp_s = a[30:23];             	// extract the exponent
	assign exp_d = {3'b000, exp_s} + 896;	// rebias
	assign zero = (a == 0) ? 1 : 0;

	assign result[63] = a[31];					// sign bit
   // biased exponent
	assign result[62:52] = (zero == 0) ? exp_d : 11'b00000000000;
	assign result[51:29] = a[22:0];  // fraction
	assign result[28:0] = 29'b00000000000000000000000000000;
endmodule
  
module dtof_ll(clk, a, result);
	input clk;
	input [64-1:0] a;
	output [32-1:0] result;

	wire [7:0] exp_s;
	wire [10:0] exp_d;
	wire [10:0] exp;
	wire ovr, inf;
	
	assign exp_d = a[62:52];             // extract the exponent
	assign exp = exp_d - 896;  // rebias
	assign exp_s = exp[7:0];
	assign ovr = (exp[10:8] == 0) ? 0 : 1;
	assign inf = (ovr == 1) ? a[62] : 0;

  	// sign bit
	assign result[31] = a[63];
  	// exponent
	assign result[30:23] = (ovr == 0) ? exp_s : 
								  ( (inf == 0) ? "00000000" : "11111111");
  	// fraction
	assign result[22:0] = (ovr == 0) ? a[51:29] : 23'b00000000000000000000000;
endmodule

module latency_delay(clk, a, a_dly, pipeEn);
	parameter WIDTH = 1;
 	parameter LATENCY = 1;

	input clk;
	input [WIDTH-1:0] a;
	output [WIDTH-1:0] a_dly;
	input pipeEn;

	reg [WIDTH-1:0] a_dly_arr [LATENCY-1:0];
	
	genvar i;
	
	generate
		for(i=0; i<LATENCY; i=i+1)
		begin
         always @(posedge clk)
         begin
				if (pipeEn)
				begin
					if (i==0)
						a_dly_arr[i] <= a;
					else
						a_dly_arr[i] <= a_dly_arr[i-1];
				end
         end
		end
	endgenerate
    
	assign a_dly = a_dly_arr[LATENCY];
endmodule

