# NOTE: Output to stdout during load OR ANY Tcl error may cause errors in impulse_sm (or Stage master Explorer) 
#			and it will spin and never exit (or sometimes bring up the Stage Master Explorer window) 
namespace eval ::xilinx::floating_point {
	namespace export hdl*
	namespace export gen*
	namespace export convert*
	
	namespace eval constants {
		#####
		# namespace constants
		variable PSPCOMMON_PATH [file join $::env(IMPULSEC_HOME) Architectures/PSPcommon/Tcl]
		variable COREGENERATOR_COM_PATH [file join $::env(IMPULSEC_HOME) Architectures/Xilinx/AutoIP/Common/COREGenerator]
		variable COREGENERATOR_FP_PATH [file join $::env(IMPULSEC_HOME) Architectures/Xilinx/AutoIP/FloatingPoint/COREGenerator]
	
		# Map XML paramter values to XCO paramter values
		# <xml param name> { name <xml param name> [<xml val> <xco val>...]} 
		variable XML_TO_XCO_VALMAP
		array set XML_TO_XCO_VALMAP {
			latency {
				name c_latency
			}
			rate {
				name c_rate
			}
			dsp {
				name c_mult_usage 
				0 No_Usage
				1 Medium_Usage
				2 Full_Usage
				3 Max_Usage
			}
			fast {
				name c_optimization
    			0 Low_Latency
    			1 Speed_Optimized
			}
			fxpt_iwidth {
				name c_a_exponent_width
			}
			fxpt_uwidth {
				name c_a_exponent_width
				expression "[calcWidth a paramArr]"
			}
			fxpt_owidth {
				name c_result_exponent_width
			}
		}
		# Map XML paramter values to simulation generic paramter values
		# <xml param name> { name <xml param name> [<xml val> <generic val>...]} 
		variable XML_TO_SIMGENERIC_VALMAP
		array set XML_TO_SIMGENERIC_VALMAP {
			latency {
				name c_latency
			}
			rate {
				name c_rate
			}
			dsp {
				name c_mult_usage 
			}
			fast {
				name c_optimization
				0 2
				1 1
			}
			fxpt_iwidth {
				name c_a_width
			}
			fxpt_uwidth {
				name c_a_width
				expression "[calcWidth a paramArr]"
			}
			fxpt_owidth {
				name c_result_width
			}
		}
		
        # Note: "float" can mean either "float 32-bit" or "double 64-bit" for IO
		# NOTE: expressions may use local variables that must be present in resolving proc
		#		example: $paramArr(width) uses array 'paramArr' local to proc calling 'subst'
        # HDL functions for 2-float-in-1-float-out modules
		namespace eval twoFloatInOneFloatOut {
			# <name> <width> [init]
			variable DEF_GENERICS {
			}
			# <name> <width> [type]
			variable DEF_SIGNALS {
			}
			# <port name> <width expr>   
			variable DEF_PORTS {
    			{"a" $paramArr(width)}
    			{"b" $paramArr(width)}
    			{"go" 0}
    			{"pipeEn" 0}
    			{"result" $paramArr(width)}
			}
			# <port name> <0=in/1=out> <width 0=signal, n>0=vector> [<signal type>]
			variable DEF_COMP_PORTS {
				{clk 0 0 std_logic}
				{a 0 $paramArr(width) std_logic_vector}
				{b 0 $paramArr(width) std_logic_vector}
				{ce 0 0 std_logic}
				{operation_nd 0 0 std_logic}
				{result 1 $paramArr(width) std_logic_vector}
			}
			# <comp port> <signal name>
			variable COMP_INST_PORT_MAP {
				clk clk
				a a 
				b b
				ce pipeEn
				operation_nd go 
				result result
			}
		}
		# HDL functions for 1-float-in-1-float-out modules
		namespace eval oneFloatInOneFloatOut {
			# <name> <width> [init]
			variable DEF_GENERICS {
			}
			# <name> <width> [type]
			variable DEF_SIGNALS {
			}
			# <port name> <width expr>
			variable DEF_PORTS {
				{"a" $paramArr(width)}
				{"go" 0}
				{"pipeEn" 0}
				{"result" $paramArr(width)}
			}
			# <port name> <0=in/1=out> <width 0=signal, n>0=vector> [<signal type>]
			variable DEF_COMP_PORTS {
				{clk 0 0 std_logic}
				{a 0 $paramArr(width) std_logic_vector}
				{ce 0 0 std_logic}
				{operation_nd 0 0 std_logic}
				{result 1 $paramArr(width) std_logic_vector}
			}
			# <comp port> <signal name>
			variable COMP_INST_PORT_MAP {
				clk clk
				a a 
				ce pipeEn
				operation_nd go 
				result result
			}
		}
		# HDL functions for 2-float-in-1-bit-out modules (e.g. comparisons)
		namespace eval twoFloatInOneBitOut {
			# <name> <width> [init]
			variable DEF_GENERICS {
			}
			# <name> <width> [type]
			variable DEF_SIGNALS {
			}
			# <port name> <width expr>
			variable DEF_PORTS {
				{"a" $paramArr(width)}
				{"b" $paramArr(width)}
				{"go" 0}
				{"pipeEn" 0}
				{"result" 1}
			}
			# <port name> <0=in/1=out> <width 0=signal, n>0=vector> [<signal type>]
			variable DEF_COMP_PORTS {
				{clk 0 0 std_logic}
				{a 0 $paramArr(width) std_logic_vector}
				{b 0 $paramArr(width) std_logic_vector}
				{ce 0 0 std_logic}
				{operation_nd 0 0 std_logic}
				{result 1 1 std_logic_vector}
			}
			# <comp port> <signal name>
			variable COMP_INST_PORT_MAP {
				clk clk
				a a 
				b b
				ce pipeEn
				operation_nd go 
				result result
			}
		}
		# HDL functions for 1-float-in-1-int-out modules (e.g. float-to-int)
		namespace eval oneFloatInOneIntOut {
			# <name> <width> [init]
			variable DEF_GENERICS {
			}
			# <name> <width> [type]
			variable DEF_SIGNALS {
			}
			# <port name> <width expr>
			variable DEF_PORTS {
				{"a" $paramArr(width)}
				{"go" 0}
				{"pipeEn" 0}
				{"result" $paramArr(fxpt_owidth)}
			}
			# <port name> <0=in/1=out> <width 0=signal, n>0=vector> [<signal type>]
			variable DEF_COMP_PORTS {
				{clk 0 0 std_logic}
				{a 0 $paramArr(width) std_logic_vector}
				{ce 0 0 std_logic}
				{operation_nd 0 0 std_logic}
				{result 1 $paramArr(fxpt_owidth) std_logic_vector}
			}
			# <comp port> <signal name>
			variable COMP_INST_PORT_MAP {
				clk clk
				a a 
				ce pipeEn
				operation_nd go 
				result result
			}
		}
        # HDL functions for 1-int-in-1-float-out modules (e.g. int-to-float)
        namespace eval oneIntInOneFloatOut {
			# <name> <width> [init]
			variable DEF_GENERICS {
			}
			# <name> <width> [type]
			variable DEF_SIGNALS {
			}
			# <port name> <width expr>
        	variable DEF_PORTS {
        		{"a" $paramArr(fxpt_iwidth)}
        		{"go" 0}
        		{"pipeEn" 0}
        		{"result" $paramArr(width)}
        	}
			# <port name> <0=in/1=out> <width 0=signal, n>0=vector> [<signal type>]
        	variable DEF_COMP_PORTS {
        		{clk 0 0 std_logic}
        		{a 0 $paramArr(fxpt_iwidth) std_logic_vector}
        		{ce 0 0 std_logic}
        		{operation_nd 0 0 std_logic}
        		{result 1 $paramArr(width) std_logic_vector}
        	}
			# <comp port> <signal name>
        	variable COMP_INST_PORT_MAP {
        		clk clk
        		a a 
        		ce pipeEn
        		operation_nd go 
        		result result
        	}
        }
		# HDL functions for 1-unsignedInt-in-1-float-out modules (e.g. int-to-float)
		namespace eval oneUIntInOneFloatOut {
			namespace export *
			# <name> <width> [init]
			variable DEF_GENERICS {
			}
			# <name> <width> [type]
			variable DEF_SIGNALS {
				{"core_result" $paramArr(width)} 
			}
			# <port name> <width expr>
			variable DEF_PORTS {
				{"a" $paramArr(fxpt_uwidth)}
				{"go" 0}
				{"pipeEn" 0}
				{"result" $paramArr(width)}
			}
			# <port name> <0=in/1=out> <width 0=signal, n>0=vector> [<signal type>]
			variable DEF_COMP_PORTS {
				{clk 0 0 std_logic}
				{a 0 [calcWidth a paramArr] std_logic_vector}
				{ce 0 0 std_logic}
				{operation_nd 0 0 std_logic}
				{result 1 $paramArr(width) std_logic_vector}
			}
			proc calcWidth {sigName paramArrIn} {
				upvar $paramArrIn paramArr
				switch $sigName {
					{a} {
        				if {$paramArr(fxpt_uwidth) < 64} {
        					return [expr $paramArr(fxpt_uwidth) + 1]
        				} else {
        					return 64
        				}
					}
					default {
						::xilinx::floating_point::fatalError "calcWidth:Don't know how to calculate width for \[$sigName\]"
					}
				}
			}
			# <comp port> <signal name>
			variable COMP_INST_PORT_MAP {
				clk clk
				a [calcCompInstPortWidth a paramArr] 
				ce pipeEn
				operation_nd go 
				result core_result
			}
			proc calcCompInstPortWidth {sigName paramArrIn} {
				upvar $paramArrIn paramArr
				switch $sigName {
					{a} {
						if {$paramArr(fxpt_uwidth) < 64} {
							#return {"'0' & a"}
							return "alt64unsigned"
						} else {
							return "a64unsigned"
						}
					}
					default {
						::xilinx::floating_point::fatalError "calcCompInstPortWidth:Don't know how to calculate width for \[$sigName\]"
					}
				}
			}
			# routine specific to this opIOType for generatign additional HDL
			# utof/utod functions
			proc genIpForOpIOType {fd opName opIOType lang compName paramArrIn} {
				upvar $paramArrIn paramArr
				
				# utod must be scaled if input is 64 bits because utod core only accepts signed
				#  integers.  If input width is < 64 then do nothing, at 64 bits scale input/output
				#  by 2 only if the MSb is set.  >64 bits is unsupported 
				set awidth $paramArr(fxpt_uwidth)
				if {$awidth > 64} {
					::xilinx::floating_point::fatalError "$opName of >64 bits is not supported, use correct type or cast to <=64 bit"
				} elseif {$awidth < 64} {
					# <64 bit input, just assign
					hdl::defineSig "alt64unsigned" [calcWidth a paramArr]
					hdl::assign "alt64unsigned" [hdl::cat [hdl::bit 0] [hdl::val "a"]]
					hdl::assign "result" [hdl::val "core_result"]
				} else {
					# 64-bit input, need to instantiate logic that does scale by 2 based on MSb
					# additional signals
					hdl::defineSig "a64unsigned" [calcWidth a paramArr]
					hdl::defineSig "msb_dly" 1
					hdl::defineSig "msb" 1
					hdl::defineSig "newexp" 11
										
					# assignments based on MSb
					hdl::assign "msb" [hdl::subvec "a" [expr $awidth - 1] [expr $awidth - 1]]
					
					hdl::assign "a64unsigned" [hdl::cat [hdl::bit 0] [hdl::subvec "a" [expr $awidth - 1] 1]] [hdl::eq [hdl::elm "a" [expr $awidth - 1]] [hdl::bit 1]] 
					hdl::assign "a64unsigned" [hdl::val "a"]
					
					hdl::assign "newexp" [hdl::add [hdl::subvec "core_result" 62 52] [hdl::cat [hdl::const "0000000000"] [hdl::val msb_dly]]]
					
					hdl::assign "result" [hdl::cat [hdl::bit 0] [hdl::cat [hdl::val "newexp"] [hdl::subvec "core_result" 51 0]]] 
					
					#hdl::assign "result" [hdl::cat "0" [hdl::cat "std_ulogic_vector(expxtwo)" "std_ulogic_vector(result_slv(51 downto 0))"]]
					
					# delay element component
					hdl::defineLibrary "impulse"
					hdl::defineComponent latency_delay [list [list LATENCY 4 "0001"]\
					 										[list WIDTH 8 "00000001"]]\
										 [list [list clk 0 0]\
										  	   [list a 0 1]\
										       [list pipeEn 0 0]\
										  	   [list a_dly 1 1]]
					# delay element instance
					hdl::instance latency_delay_inst latency_delay {}\
									[list clk clk a msb pipeEn pipeEn a_dly msb_dly]
				}
			}
		}
		# import constants into caller's space as local variables with same names
		proc import {args} {
            set nmToImport [linsert $args 0 ""]
            foreach {subConstNS} $nmToImport {
               	if {$subConstNS != ""} {
                	set subConstNS "::${subConstNS}"
                }
                #puts "importing sub-constant namespace \[$subConstNS\]"
                foreach {const} [info vars "[namespace current]${subConstNS}::*"] {
                #puts "importing $const"
                	uplevel 1 [list variable "$const"]
                }
            }
		}
	}
	
	#load dependency Tcl
	proc init {} {
		source [file join $::xilinx::floating_point::constants::PSPCOMMON_PATH ic_proj.tcl]
		source [file join $::xilinx::floating_point::constants::COREGENERATOR_COM_PATH xco.tcl]
		source [file join $::xilinx::floating_point::constants::COREGENERATOR_COM_PATH xcoSim.tcl]
	}

	proc hdlDefineGenerics {paramArrIn defGenericsList} {
		upvar $paramArrIn paramArr
		foreach {genericDef} $defGenericsList {
			puts "Defining generic $genericDef"
			eval hdl::defineGeneric [join $genericDef]
		}
	}

	proc hdlDefineSignals {paramArrIn defSignalsList} {
		upvar $paramArrIn paramArr
		foreach {signalDef} $defSignalsList {
			eval hdl::defineSig [join $signalDef]
		}
	}
	
	proc hdlDefinePorts {paramArrIn defPortList} {
		upvar $paramArrIn paramArr
		foreach {portDef} $defPortList {
			eval hdl::definePort [join $portDef]
		}
	}
	
	proc hdlDefineComponent {compName paramArrIn defCompPortsList} {
		upvar $paramArrIn paramArr
		set generics { }
		set ports {}
		foreach {portDef} $defCompPortsList {
			lappend ports [subst $portDef]
		}
		hdl::defineComponent $compName $generics $ports 
	}
	
	proc hdlInstanceComponent {compInstName paramArrIn portMapList} {
		upvar $paramArrIn paramArr
		set substList [subst $portMapList]
		#puts "substList=$substList"
		hdl::instance "${compInstName}_inst" $compInstName {} $substList   
	}
	
	proc fatalError {errStr} {
		puts ""
		puts "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!"
		puts "!! FATAL_ERROR:"
		puts "!!   $errStr"
		puts "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!"
		puts "Exiting on fatal error..."
		exit -1;
	}
	
	# apply translation array on parameter list
	proc applyTranslationArray { paramList trasnArrIn} {
		upvar $trasnArrIn trasnArr
		array set paramArr $paramList
		array set outParams {}
		foreach {key val} $paramList {
			#puts "mapping $key $val"
			if {[info exists trasnArr($key)]} {
				array set xmlVal2XcoArr $trasnArr($key)
				set xcoKey $xmlVal2XcoArr(name) 
				if {[info exists xmlVal2XcoArr($val)]} {
					set xcoVal $xmlVal2XcoArr($val)
				} elseif {[info exists xmlVal2XcoArr(expression)]} {
					set xcoVal [subst $xmlVal2XcoArr(expression)]
				} else {
					set xcoVal $val
				}
				# unset for clean next lookup
				unset xmlVal2XcoArr
				set outParams($xcoKey) $xcoVal
			}
		}
		return [array get outParams]
	}
	
	# lookup, and translate if necessary, XML parameters into XCO 
	proc convertXmlToXcoParameters {xmlParamList} {
		constants::import
		return [applyTranslationArray $xmlParamList XML_TO_XCO_VALMAP]
	}

	# lookup, and translate if necessary, XML parameters into generics 
	proc convertXmlToGenericParameters {xmlParamList} {
		constants::import
		return [applyTranslationArray $xmlParamList XML_TO_SIMGENERIC_VALMAP]
	}
	
	# convert the DEF_COMP_PORTS into a list ready to become an array of <port name>=<list>
	#  also resolves values in <list>
	proc convertCompPortListToArrayList {paramArrIn defCompPortsList} {
		upvar $paramArrIn paramArr
		set arrList [list]
		foreach {portDef} $defCompPortsList {
			set resolved [subst $portDef]
			lappend arrList [lindex $resolved 0] [lrange $resolved 1 end]
		}
#		puts "arrList=$arrList"
		return $arrList
	}

	proc printComment {fd lang comment} {
		if {$lang == "vhdl"} {
			set commTxt "-- "
		} else {
			set commTxt "// "
		}
		puts $fd "${commTxt}${comment}"
	}
	
	proc genHdlForOp {fd opName opIOType lang compName paramList} {
		# bring specific configuration variables for specified op's type into scope
		constants::import $opIOType
		namespace import constants::${opIOType}::* 

		array set paramArr $paramList
		set xcoCompName "${compName}_coregen"
		
		# add ports to HDL to be generated
		hdlDefinePorts paramArr $DEF_PORTS
		hdlDefineGenerics paramArr $DEF_GENERICS
		hdlDefineComponent $xcoCompName paramArr $DEF_COMP_PORTS
		hdlDefineSignals paramArr $DEF_SIGNALS
		
		# add body to HDL to be generated
		hdlInstanceComponent $xcoCompName paramArr $COMP_INST_PORT_MAP
		
		# call additional logic generation if routine exists
		if {[info proc genIpForOpIOType] != ""} { 
			genIpForOpIOType $fd $opName $opIOType $lang $compName paramArr
		}
		
		# write HDL for autoip function
		printComment $fd $lang "$compName is instantiated within '_com.vhd file"
		hdl::gen $fd $compName $lang

		# generate specialized XCO file from template file
		array set xcoOutputParamsArr {}
		array set replaceParmsArr [convertXmlToXcoParameters $paramList]
		set replaceParmsArr(component_name) $xcoCompName
		if {$lang != "vhdl"} {
			# XCO files are VHDL-centric by default, change some settings for Verilog
			set replaceParmsArr(designentry) "Verilog"
			set replaceParmsArr(simulationfiles) "Structural"
			set replaceParmsArr(verilogsim) "True"
			set replaceParmsArr(vhdlsim) "False"
		}
		set inXcoFileName [file join $COREGENERATOR_FP_PATH "${opName}.xco"]
		set outXcoDirName [file join [::ic_proj::getParmVal GenCodeHWTargetDir] xco]
		if {[file exists $outXcoDirName] == 0} {
			file mkdir $outXcoDirName
		}
		set outXcoFileName [file join $outXcoDirName "$xcoCompName.xco" ]
		::xilinx::xco::setParams $inXcoFileName $outXcoFileName replaceParmsArr xcoOutputParamsArr

		# write HDL simulation module that uses XilinxCoreLib (avoids having to run CORE Generator)
		#  uses XCO output parameters to set generics.  ONLY wokrs for VHDL, Verilog requires CORE
		#  Generator to be run in order to create .v simulation files
		if {$lang == "vhdl"} {
   		set inXcoSimeFileName [file join $COREGENERATOR_FP_PATH "${opName}.xcoSim"]
   		# translate XCO paramters to VHDL generics
   		array set replaceGenericsArr [convertXmlToGenericParameters $paramList]
   		set replaceGenericsArr(component_name) $xcoCompName 
   		array set replacePortsArr [convertCompPortListToArrayList paramArr $DEF_COMP_PORTS]
   		::xilinx::xcosim::parseAndWriteXcoSimFile $fd $inXcoSimeFileName replaceGenericsArr replacePortsArr 
		}		
		# remove imported specific configuration (next call may need to import a different set)
		namespace forget constants::${opIOType}::*
		namespace forget constants::${opIOType}::*
	}
	
	proc genIp {fd lang compName paramList} {
		puts "genIp $lang $compName $paramList"
		switch -regexp $compName {
			{fadd_+} -
			{fsub_+} -
			{fmul_+} -
			{fdiv_+} -
			{faddd_+} -
			{fsubd_+} -
			{fmuld_+} -
			{fdivd_+} {
				set opName "[lindex [split $compName "_"] 0]"
				set opIOType twoFloatInOneFloatOut
			}
			{sqrt_+} -
			{sqrtd_+} {
				set opName "[lindex [split $compName "_"] 0]"
				set opIOType oneFloatInOneFloatOut
			}
			{fcmp_+} {
				set opName "f[lindex [split $compName "_"] 1]"
				set opIOType twoFloatInOneBitOut
			}
			{fcmpd_+} {
				set opName "f[lindex [split $compName "_"] 1]d"
				set opIOType twoFloatInOneBitOut
			}
			{[fd]toi_+} {
				set opName "[lindex [split $compName "_"] 0]"
				set opIOType oneFloatInOneIntOut
			}
			{ito[fd]_+} {
				set opName "[lindex [split $compName "_"] 0]"
				set opIOType oneIntInOneFloatOut
			}
			{uto[fd]_+} {
				set opName "[lindex [split $compName "_"] 0]"
				set opIOType oneUIntInOneFloatOut
			}
			default {
				fatalError "Unknown component name $compName"
			}
		}
		genHdlForOp $fd $opName $opIOType $lang $compName $paramList
	}
	
	proc calcTiming {opName field paramList} {
		switch $field {
			rate {
				# return 1
				array set paramArr $paramList
				return $paramArr(rate)
			}
			cycles {
				array set paramArr $paramList
				return $paramArr(latency)
			}
		}
	}
	
	#####
	# initialize library on load 
	init
	
}

return 0